#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""The core API for interacting with Discord directly."""

from hikari.core.clients import app_config
from hikari.core.clients import bot_client
from hikari.core.clients import gateway_client
from hikari.core.clients import gateway_config
from hikari.core.clients import http_client
from hikari.core.clients import http_config
from hikari.core.clients import protocol_config
from hikari.core.clients import shard_client

from hikari.core.clients.app_config import *
from hikari.core.clients.bot_client import *
from hikari.core.clients.gateway_client import *
from hikari.core.clients.gateway_config import *
from hikari.core.clients.http_client import *
from hikari.core.clients.http_config import *
from hikari.core.clients.protocol_config import *
from hikari.core.clients.shard_client import *


__all__ = [
    *app_config.__all__,
    *bot_client.__all__,
    *gateway_client.__all__,
    *gateway_config.__all__,
    *http_client.__all__,
    *http_config.__all__,
    *protocol_config.__all__,
    *shard_client.__all__,
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe webhooks on Discord."""
__all__ = ["WebhookType", "Webhook"]

import enum
import typing

from hikari.core import entities
from hikari.core import snowflakes
from hikari.core import users
from hikari.internal_utilities import marshaller


@enum.unique
class WebhookType(enum.IntEnum):
    #: Incoming webhook.
    INCOMING = 1
    #: Channel Follower webhook.
    CHANNEL_FOLLOWER = 2


@marshaller.attrs(slots=True)
class Webhook(snowflakes.UniqueEntity, entities.Deserializable):
    """Represents a webhook"""

    #: The type of the webhook.
    #:
    #: :type: :obj:`WebhookType`
    type: WebhookType = marshaller.attrib(deserializer=WebhookType)

    #: The guild ID of the webhook.
    #:
    #: :type: :obj:`snowflakes.Snowflake`, optional
    guild_id: typing.Optional[snowflakes.Snowflake] = marshaller.attrib(
        deserializer=snowflakes.Snowflake.deserialize, if_undefined=None
    )

    #: The channel ID this webhook is for.
    #:
    #: :type: :obj:`snowflakes.Snowflake`
    channel_id: snowflakes.Snowflake = marshaller.attrib(deserializer=snowflakes.Snowflake.deserialize)

    #: The user that created the webhook
    #:
    #: Note
    #: ----
    #: This will be ``None`` when getting a webhook with a token
    #:
    #: :type: :obj:`users.User`, optional
    user: typing.Optional[users.User] = marshaller.attrib(deserializer=users.User.deserialize, if_undefined=None)

    #: The default name of the webhook.
    #:
    #: :type: :obj:`str`, optional
    name: typing.Optional[str] = marshaller.attrib(deserializer=str, if_none=None)

    #: The default avatar hash of the webhook.
    #:
    #: :type: :obj:`str`, optional
    avatar_hash: typing.Optional[str] = marshaller.attrib(raw_name="avatar", deserializer=str, if_none=None)

    #: The token of the webhook.
    #:
    #: Note
    #: ----
    #: This is only available for Incoming webhooks.
    #:
    #: :type: :obj:`str`, optional
    token: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe voice states on Discord.
"""
__all__ = ["VoiceRegion", "VoiceState"]

import typing

from hikari.core import entities
from hikari.core import guilds
from hikari.core import snowflakes
from hikari.internal_utilities import marshaller


@marshaller.attrs(slots=True)
class VoiceState(entities.HikariEntity, entities.Deserializable):
    """Represents a user's voice connection status."""

    #: The ID of the guild this voice state is in, if applicable.
    #:
    #: :type: :obj:`snowflakes.Snowflake`, optional
    guild_id: typing.Optional[snowflakes.Snowflake] = marshaller.attrib(
        deserializer=snowflakes.Snowflake.deserialize, if_undefined=None
    )

    #: The ID of the channel this user is connected to.
    #:
    #: :type: :obj:`.core.snowflakes.Snowflake`, optional
    channel_id: snowflakes.Snowflake = marshaller.attrib(deserializer=snowflakes.Snowflake.deserialize)

    #: The ID of the user this voice state is for.
    #:
    #: :type: :obj:`snowflakes.Snowflake`
    user_id: snowflakes.Snowflake = marshaller.attrib(deserializer=snowflakes.Snowflake.deserialize)

    #: The guild member this voice state is for if the voice state is in a
    #: guild.
    #:
    #: :type: :obj:`guilds.GuildMember`, optional
    member: typing.Optional[guilds.GuildMember] = marshaller.attrib(
        deserializer=guilds.GuildMember.deserialize, if_undefined=None
    )

    #: The ID of this voice state's session.
    #:
    #: :type: :obj:`str`
    session_id: str = marshaller.attrib(deserializer=str)

    #: Whether this user is deafened by the guild.
    #:
    #: :type: :obj:`bool`
    is_guild_deafened: bool = marshaller.attrib(raw_name="deaf", deserializer=bool)

    #: Whether this user is muted by the guild.
    #:
    #: :type: :obj:`bool`
    is_guild_muted: bool = marshaller.attrib(raw_name="mute", deserializer=bool)

    #: Whether this user is deafened by their client.
    #:
    #: :type: :obj:`bool`
    is_self_deafened: bool = marshaller.attrib(raw_name="self_deaf", deserializer=bool)

    #: Whether this user is muted by their client.
    #:
    #: :type: :obj:`bool`
    is_self_muted: bool = marshaller.attrib(raw_name="self_mute", deserializer=bool)

    #: Whether this user is streaming using "Go Live".
    #:
    #: :type: :obj:`bool`
    is_streaming: bool = marshaller.attrib(raw_name="self_stream", deserializer=bool, if_undefined=False)

    #: Whether this user is muted by the current user.
    #:
    #: :type: :obj:`bool`
    is_suppressed: bool = marshaller.attrib(raw_name="suppress", deserializer=bool)


@marshaller.attrs(slots=True)
class VoiceRegion(entities.HikariEntity, entities.Deserializable):
    """Represent's a voice region server."""

    #: The ID of this region
    #:
    #: :type: :obj:`str`
    id: str = marshaller.attrib(deserializer=str)

    #: The name of this region
    #:
    #: :type: :obj:`str`
    name: str = marshaller.attrib(deserializer=str)

    #: Whether this region is vip-only.
    #:
    #: :type: :obj:`bool`
    is_vip: bool = marshaller.attrib(raw_name="vip", deserializer=bool)

    #: Whether this region's server is closest to the current user's client.
    #:
    #: :type: :obj:`bool`
    is_optimal_location: bool = marshaller.attrib(raw_name="optimal", deserializer=bool)

    #: Whether this region is deprecated.
    #:
    #: :type: :obj:`bool`
    is_deprecated: bool = marshaller.attrib(raw_name="deprecated", deserializer=bool)

    #: Whether this region is custom (e.g. used for events).
    #:
    #: :type: :obj:`bool`
    is_custom: bool = marshaller.attrib(raw_name="custom", deserializer=bool)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe guilds on Discord.
"""
__all__ = [
    "ActivityFlag",
    "ActivityType",
    "GuildChannel",
    "GuildTextChannel",
    "GuildNewsChannel",
    "GuildStoreChannel",
    "GuildVoiceChannel",
    "GuildCategory",
    "GuildRole",
    "GuildFeature",
    "GuildSystemChannelFlag",
    "GuildMessageNotificationsLevel",
    "GuildExplicitContentFilterLevel",
    "GuildMFALevel",
    "GuildVerificationLevel",
    "GuildPremiumTier",
    "Guild",
    "GuildMember",
    "GuildMemberPresence",
    "GuildIntegration",
    "GuildMemberBan",
    "PartialGuild",
    "PresenceStatus",
]

import datetime
import enum
import typing

from hikari.core import channels
from hikari.core import colors
from hikari.core import emojis as _emojis
from hikari.core import entities
from hikari.core import permissions as _permissions
from hikari.core import snowflakes
from hikari.core import users
from hikari.internal_utilities import cdn
from hikari.internal_utilities import dates
from hikari.internal_utilities import marshaller
from hikari.internal_utilities import transformations


@marshaller.attrs(slots=True)
class GuildChannel(channels.Channel, entities.Deserializable):
    """The base for anything that is a guild channel."""


@marshaller.attrs(slots=True)
class GuildTextChannel(GuildChannel):
    ...


@marshaller.attrs(slots=True)
class GuildVoiceChannel(GuildChannel):
    ...


@marshaller.attrs(slots=True)
class GuildCategory(GuildChannel):
    ...


@marshaller.attrs(slots=True)
class GuildStoreChannel(GuildChannel):
    ...


@marshaller.attrs(slots=True)
class GuildNewsChannel(GuildChannel):
    ...


def parse_guild_channel(payload) -> GuildChannel:
    class Duff:
        id = snowflakes.Snowflake(123)

    # FIXME: implement properly
    return Duff()


@enum.unique
class GuildExplicitContentFilterLevel(enum.IntEnum):
    """Represents the explicit content filter setting for a guild."""

    #: No explicit content filter.
    DISABLED = 0

    #: Filter posts from anyone without a role.
    MEMBERS_WITHOUT_ROLES = 1

    #: Filter all posts.
    ALL_MEMBERS = 2


@enum.unique
class GuildFeature(str, enum.Enum):
    """Features that a guild can provide."""

    #: Guild has access to set an animated guild icon.
    ANIMATED_ICON = "ANIMATED_ICON"
    #: Guild has access to set a guild banner image.
    BANNER = "BANNER"
    #: Guild has access to use commerce features (i.e. create store channels).
    COMMERCE = "COMMERCE"
    #: Guild is able to be discovered in the directory.
    DISCOVERABLE = "DISCOVERABLE"
    #: Guild is able to be featured in the directory.
    FEATURABLE = "FEATURABLE"
    #: Guild has access to set an invite splash background.
    INVITE_SPLASH = "INVITE_SPLASH"
    #: More emojis can be hosted in this guild than normal.
    MORE_EMOJI = "MORE_EMOJI"
    #: Guild has access to create news channels.
    NEWS = "NEWS"
    #: People can view channels in this guild without joining.
    LURKABLE = "LURKABLE"
    #: Guild is partnered.
    PARTNERED = "PARTNERED"
    #: Guild is public, go figure.
    PUBLIC = "PUBLIC"
    #: Guild cannot be public. Who would have guessed?
    PUBLIC_DISABLED = "PUBLIC_DISABLED"
    #: Guild has access to set a vanity URL.
    VANITY_URL = "VANITY_URL"
    #: Guild is verified.
    VERIFIED = "VERIFIED"
    #: Guild has access to set 384kbps bitrate in voice (previously
    #: VIP voice servers).
    VIP_REGIONS = "VIP_REGIONS"


@enum.unique
class GuildMessageNotificationsLevel(enum.IntEnum):
    """Represents the default notification level for new messages in a guild."""

    #: Notify users when any message is sent.
    ALL_MESSAGES = 0

    #: Only notify users when they are @mentioned.
    ONLY_MENTIONS = 1


@enum.unique
class GuildMFALevel(enum.IntEnum):
    """Represents the multi-factor authorization requirement for a guild."""

    #: No MFA requirement.
    NONE = 0

    #: MFA requirement.
    ELEVATED = 1


@enum.unique
class GuildPremiumTier(enum.IntEnum):
    """Tier for Discord Nitro boosting in a guild."""

    #: No Nitro boosts.
    NONE = 0

    #: Level 1 Nitro boost.
    TIER_1 = 1

    #: Level 2 Nitro boost.
    TIER_2 = 2

    #: Level 3 Nitro boost.
    TIER_3 = 3


class GuildSystemChannelFlag(enum.IntFlag):
    """Defines which features are suppressed in the system channel."""

    #: Display a message about new users joining.
    SUPPRESS_USER_JOIN = 1
    #: Display a message when the guild is Nitro boosted.
    SUPPRESS_PREMIUM_SUBSCRIPTION = 2


@enum.unique
class GuildVerificationLevel(enum.IntEnum):
    """Represents the level of verification a user needs to provide for their
    account before being allowed to participate in a guild."""

    #: Unrestricted
    NONE = 0

    #: Must have a verified email on their account.
    LOW = 1

    #: Must have been registered on Discord for more than 5 minutes.
    MEDIUM = 2

    #: (╯°□°）╯︵ ┻━┻ - must be a member of the guild for longer than 10 minutes.
    HIGH = 3

    #: ┻━┻ミヽ(ಠ益ಠ)ﾉ彡┻━┻ - must have a verified phone number.
    VERY_HIGH = 4


@marshaller.attrs(slots=True)
class GuildMember(entities.HikariEntity, entities.Deserializable):
    """Used to represent a guild bound member."""

    #: This member's user object.
    #:
    #: :type: :obj:`users.User`
    user: users.User = marshaller.attrib(deserializer=users.User.deserialize)

    #: This member's nickname, if set.
    #:
    #: :type: :obj:`str`, optional
    nickname: typing.Optional[str] = marshaller.attrib(raw_name="nick", deserializer=str, if_none=None)

    #: A sequence of the IDs of the member's current roles.
    #:
    #: :type: :obj:`typing.Sequence` [ :obj:`snowflakes.Snowflake` ]
    role_ids: typing.Sequence[snowflakes.Snowflake] = marshaller.attrib(
        raw_name="roles", deserializer=lambda role_ids: [snowflakes.Snowflake.deserialize(rid) for rid in role_ids],
    )

    #: The datetime of when this member joined the guild they belong to.
    #:
    #: :type: :obj:`datetime.datetime`
    joined_at: datetime.datetime = marshaller.attrib(deserializer=dates.parse_iso_8601_ts)

    #: The datetime of when this member started "boosting" this guild.
    #: Will be ``None`` if they aren't boosting.
    #:
    #: :type: :obj:`datetime.datetime`, optional
    premium_since: typing.Optional[datetime.datetime] = marshaller.attrib(
        deserializer=dates.parse_iso_8601_ts, if_none=None, if_undefined=None,
    )

    #: Whether this member is deafened by this guild in it's voice channels.
    #:
    #: :type: :obj:`bool`
    is_deaf: bool = marshaller.attrib(raw_name="deaf", deserializer=bool)

    #: Whether this member is muted by this guild in it's voice channels.
    #:
    #: :type: :obj:`bool`
    is_mute: bool = marshaller.attrib(raw_name="mute", deserializer=bool)


@marshaller.attrs(slots=True)
class GuildRole(snowflakes.UniqueEntity, entities.Deserializable):
    """Represents a guild bound Role object."""

    #: The role's name
    #:
    #: :type: :obj:`str`
    name: str = marshaller.attrib(deserializer=str)

    #: The colour of this role, will be applied to a member's name in chat
    #: if it's their top coloured role.
    #:
    #: :type: :obj:`colors.Color`
    color: colors.Color = marshaller.attrib(deserializer=colors.Color)

    #: Whether this role is hoisting the members it's attached to in the member
    #: list, members will be hoisted under their highest role where
    #: :attr:`hoisted` is true.
    #:
    #: :type: :obj:`bool`
    is_hoisted: bool = marshaller.attrib(raw_name="hoist", deserializer=bool)

    #: The position of this role in the role hierarchy.
    #:
    #: :type: :obj:`int`
    position: int = marshaller.attrib(deserializer=int)

    #: The guild wide permissions this role gives to the members it's attached
    #: to, may be overridden by channel overwrites.
    #:
    #: :type: :obj:`_permissions.Permission`
    permissions: _permissions.Permission = marshaller.attrib(deserializer=_permissions.Permission)

    #: Whether this role is managed by an integration.
    #:
    #: :type: :obj:`bool`
    is_managed: bool = marshaller.attrib(raw_name="managed", deserializer=bool)

    #: Whether this role can be mentioned by all, regardless of the
    #: ``MENTION_EVERYONE`` permission.
    #:
    #: :type: :obj:`bool`
    is_mentionable: bool = marshaller.attrib(raw_name="mentionable", deserializer=bool)


@enum.unique
class ActivityType(enum.IntEnum):
    """
    The activity state.
    """

    #: Shows up as ``Playing <name>``
    PLAYING = 0
    #: Shows up as ``Streaming <name>``.
    #:
    #: Warnings
    #: --------
    #: Corresponding presences must be associated with VALID Twitch or YouTube
    #: stream URLS!
    STREAMING = 1
    #: Shows up as ``Listening to <name>``.
    LISTENING = 2
    #: Shows up as ``Watching <name>``. Note that this is not officially
    #: documented, so will be likely removed in the near future.
    WATCHING = 3
    #: A custom status.
    #:
    #: To set an emoji with the status, place a unicode emoji or Discord emoji
    #: (``:smiley:``) as the first part of the status activity name.
    CUSTOM = 4


@marshaller.attrs(slots=True)
class ActivityTimestamps(entities.HikariEntity, entities.Deserializable):
    """The datetimes for the start and/or end of an activity session."""

    #: When this activity's session was started, if applicable.
    #:
    #: :type: :obj:`datetime.datetime`, optional
    start: typing.Optional[datetime.datetime] = marshaller.attrib(
        deserializer=dates.unix_epoch_to_ts, if_undefined=None
    )

    #: When this activity's session will end, if applicable.
    #:
    #: :type: :obj:`datetime.datetime`, optional
    end: typing.Optional[datetime.datetime] = marshaller.attrib(deserializer=dates.unix_epoch_to_ts, if_undefined=None)


@marshaller.attrs(slots=True)
class ActivityParty(entities.HikariEntity, entities.Deserializable):
    """Used to represent activity groups of users."""

    #: The string id of this party instance, if set.
    #:
    #: :type: :obj:`str`, optional
    id: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None)

    #: The size metadata of this party, if applicable.
    #:
    #: :type: :obj:`typing.Tuple` [ :obj:`int`, :obj:`int` ], optional
    _size_information: typing.Optional[typing.Tuple[int, int]] = marshaller.attrib(
        raw_name="size", deserializer=tuple, if_undefined=None,
    )

    @property
    def current_size(self) -> typing.Optional[int]:
        """The current size of this party, if applicable."""
        return self._size_information and self._size_information[0] or None

    @property
    def max_size(self) -> typing.Optional[int]:
        """The maximum size of this party, if applicable"""
        return self._size_information and self._size_information[1] or None


@marshaller.attrs(slots=True)
class ActivityAssets(entities.HikariEntity, entities.Deserializable):
    """Used to represent possible assets for an activity."""

    #: The ID of the asset's large image, if set.
    #:
    #: :type: :obj:`str`, optional
    large_image: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None)

    #: The text that'll appear when hovering over the large image, if set.
    #:
    #: :type: :obj:`str`, optional
    large_text: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None)

    #: The ID of the asset's small image, if set.
    #:
    #: :type: :obj:`str`, optional
    small_image: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None)

    #: The text that'll appear when hovering over the small image, if set.
    #:
    #: :type: :obj:`str`, optional
    small_text: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None)


@marshaller.attrs(slots=True)
class ActivitySecret(entities.HikariEntity, entities.Deserializable):
    """The secrets used for interacting with an activity party."""

    #: The secret used for joining a party, if applicable.
    #:
    #: :type: :obj:`str`, optional
    join: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None)

    #: The secret used for spectating a party, if applicable.
    #:
    #: :type: :obj:`str`, optional
    spectate: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None)

    #: The secret used for joining a party, if applicable.
    #:
    #: :type: :obj:`str`, optional
    match: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None)


class ActivityFlag(enum.IntFlag):
    """
    Flags that describe what an activity includes,
    can be more than one using bitwise-combinations.
    """

    INSTANCE = 1 << 0
    JOIN = 1 << 1
    SPECTATE = 1 << 2
    JOIN_REQUEST = 1 << 3
    SYNC = 1 << 4
    PLAY = 1 << 5


@marshaller.attrs(slots=True)
class PresenceActivity(entities.HikariEntity, entities.Deserializable):
    """Represents an activity that'll be attached to a member's presence."""

    #: The activity's name.
    #:
    #: :type: :obj:`str`
    name: str = marshaller.attrib(deserializer=str)

    #: The activity's type.
    #:
    #: :type: :obj:`ActivityType`
    type: ActivityType = marshaller.attrib(deserializer=ActivityType)

    #: The url for a ``STREAM` type activity, if applicable
    #:
    #: :type: :obj:`url`, optional
    url: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, if_none=None)

    #: When this activity was added to the user's session.
    #:
    #: :type: :obj:`datetime.datetime`
    created_at: datetime.datetime = marshaller.attrib(deserializer=dates.unix_epoch_to_ts)

    #: The timestamps for when this activity's current state will start and
    #: end, if applicable.
    #:
    #: :type: :obj:`ActivityTimestamps`, optional
    timestamps: typing.Optional[ActivityTimestamps] = marshaller.attrib(
        deserializer=ActivityTimestamps.deserialize, if_undefined=None
    )

    #: The ID of the application this activity is for, if applicable.
    #:
    #: :type: :obj:`snowflakes.Snowflake`, optional
    application_id: typing.Optional[snowflakes.Snowflake] = marshaller.attrib(
        deserializer=snowflakes.Snowflake.deserialize, if_undefined=None
    )

    #: The text that describes what the activity's target is doing, if set.
    #:
    #: :type: :obj:`str`, optional
    details: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, if_none=None)

    #: The current status of this activity's target, if set.
    #:
    #: :type: :obj:`str`, optional
    state: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, if_none=None)

    #: The emoji of this activity, if it is a custom status and set.
    #:
    #: :type: :obj:`typing.Union` [ :obj:`_emojis.UnicodeEmoji`, :obj:`_emojis.UnknownEmoji` ], optional
    emoji: typing.Union[None, _emojis.UnicodeEmoji, _emojis.UnknownEmoji] = marshaller.attrib(
        deserializer=_emojis.deserialize_reaction_emoji, if_undefined=None
    )

    #: Information about the party associated with this activity, if set.
    #:
    #: :type: :obj:`ActivityParty`, optional
    party: typing.Optional[ActivityParty] = marshaller.attrib(deserializer=ActivityParty.deserialize, if_undefined=None)

    #: Images and their hover over text for the activity.
    #:
    #: :type: :obj:`ActivityAssets`, optional
    assets: typing.Optional[ActivityAssets] = marshaller.attrib(
        deserializer=ActivityAssets.deserialize, if_undefined=None
    )

    #: Secrets for Rich Presence joining and spectating.
    #:
    #: :type: :obj:`ActivitySecret`, optional
    secrets: typing.Optional[ActivitySecret] = marshaller.attrib(
        deserializer=ActivitySecret.deserialize, if_undefined=None
    )

    #: Whether this activity is an instanced game session.
    #:
    #: :type: :obj:`bool`, optional
    is_instance: typing.Optional[bool] = marshaller.attrib(raw_name="instance", deserializer=bool, if_undefined=None)

    #: Flags that describe what the activity includes.
    #:
    #: :type: :obj:`ActivityFlag`
    flags: ActivityFlag = marshaller.attrib(deserializer=ActivityFlag, if_undefined=None)


class PresenceStatus(enum.Enum):
    """
    The status of a member.
    """

    #: Online/green.
    ONLINE = "online"
    #: Idle/yellow.
    IDLE = "idle"
    #: Do not disturb/red.
    DND = "dnd"
    #: An alias for :attr:`DND`
    DO_NOT_DISTURB = DND
    #: Offline or invisible/grey.
    OFFLINE = "offline"


@marshaller.attrs(slots=True)
class ClientStatus(entities.HikariEntity, entities.Deserializable):
    """The client statuses for this member."""

    #: The status of the target user's desktop session.
    #:
    #: :type: :obj:`PresenceStatus`
    desktop: PresenceStatus = marshaller.attrib(
        deserializer=PresenceStatus, if_undefined=lambda: PresenceStatus.OFFLINE,
    )

    #: The status of the target user's mobile session.
    #:
    #: :type: :obj:`PresenceStatus`
    mobile: PresenceStatus = marshaller.attrib(deserializer=PresenceStatus, if_undefined=lambda: PresenceStatus.OFFLINE)

    #: The status of the target user's web session.
    #:
    #: :type: :obj:`PresenceStatus`
    web: PresenceStatus = marshaller.attrib(deserializer=PresenceStatus, if_undefined=lambda: PresenceStatus.OFFLINE)


@marshaller.attrs(slots=True)
class PresenceUser(users.User):
    """A user representation specifically used for presence updates.

    Warnings
    --------
    Every attribute except :attr:`id` may be received as :obj:`entities.UNSET`
    unless it is specifically being modified for this update.
    """

    #: This user's discriminator.
    #:
    #: :type: :obj:`typing.Union` [ :obj:`str`, `entities.UNSET` ]
    discriminator: typing.Union[str, entities.Unset] = marshaller.attrib(deserializer=str, if_undefined=entities.Unset)

    #: This user's username.
    #:
    #: :type: :obj:`typing.Union` [ :obj:`str`, `entities.UNSET` ]
    username: typing.Union[str, entities.Unset] = marshaller.attrib(deserializer=str, if_undefined=entities.Unset)

    #: This user's avatar hash, if set.
    #:
    #: :type: :obj:`typing.Union` [ :obj:`str`, `entities.UNSET` ], optional
    avatar_hash: typing.Union[None, str, entities.Unset] = marshaller.attrib(
        raw_name="avatar", deserializer=str, if_none=None, if_undefined=entities.Unset
    )

    #: Whether this user is a bot account.
    #:
    #: :type:  :obj:`typing.Union` [ :obj:`bool`, `entities.UNSET` ]
    is_bot: typing.Union[bool, entities.Unset] = marshaller.attrib(
        raw_name="bot", deserializer=bool, if_undefined=entities.Unset
    )

    #: Whether this user is a system account.
    #:
    #: :type:  :obj:`typing.Union` [ :obj:`bool`, `entities.UNSET` ]
    is_system: typing.Union[bool, entities.Unset] = marshaller.attrib(
        raw_name="system", deserializer=bool, if_undefined=entities.Unset,
    )


@marshaller.attrs(slots=True)
class GuildMemberPresence(entities.HikariEntity, entities.Deserializable):
    """Used to represent a guild member's presence."""

    #: The object of the user who this presence is for, only `id` is guaranteed
    #: for this partial object, with other attributes only being included when
    #: when they are being changed in an event.
    #:
    #: :type: :obj:`PresenceUser`
    user: PresenceUser = marshaller.attrib(deserializer=PresenceUser.deserialize)

    #: A sequence of the ids of the user's current roles in the guild this
    #: presence belongs to.
    #:
    #: :type: :obj:`typing.Sequence` [ :obj:`snowflakes.Snowflake` ]
    role_ids: typing.Sequence[snowflakes.Snowflake] = marshaller.attrib(
        raw_name="roles", deserializer=lambda roles: [snowflakes.Snowflake.deserialize(rid) for rid in roles],
    )

    #: The ID of the guild this presence belongs to.
    #:
    #: :type: :obj:`snowflakes.Snowflake`
    guild_id: snowflakes.Snowflake = marshaller.attrib(deserializer=snowflakes.Snowflake.deserialize)

    #: This user's current status being displayed by the client.
    #:
    #: :type: :obj:`PresenceStatus`
    visible_status: PresenceStatus = marshaller.attrib(raw_name="status", deserializer=PresenceStatus)

    #: An array of the user's activities, with the top one will being
    #: prioritised by the client.
    #:
    #: :type: :obj:`typing.Sequence` [ :obj:`PresenceActivity` ]
    activities: typing.Sequence[PresenceActivity] = marshaller.attrib(
        deserializer=lambda activities: [PresenceActivity.deserialize(a) for a in activities]
    )

    #: An object of the target user's client statuses.
    #:
    #: :type: :obj:`ClientStatus`
    client_status: ClientStatus = marshaller.attrib(deserializer=ClientStatus.deserialize)

    #: The datetime of when this member started "boosting" this guild.
    #: Will be ``None`` if they aren't boosting.
    #:
    #: :type: :obj:`datetime.datetime`, optional
    premium_since: typing.Optional[datetime.datetime] = marshaller.attrib(
        deserializer=dates.parse_iso_8601_ts, if_none=None, if_undefined=None,
    )

    #: This member's nickname, if set.
    #:
    #: :type: :obj:`str`, optional
    nick: typing.Optional[str] = marshaller.attrib(raw_name="nick", deserializer=str, if_undefined=None, if_none=None)


@enum.unique
class IntegrationExpireBehaviour(enum.IntEnum):
    """Behavior for expiring integration subscribers."""

    #: Remove the role.
    REMOVE_ROLE = 0
    #: Kick the subscriber.
    KICK = 1


@marshaller.attrs(slots=True)
class IntegrationAccount(entities.HikariEntity, entities.Deserializable):
    """An account that's linked to an integration."""

    #: The string ID of this (likely) third party account.
    #:
    #: :type: :obj:`str`
    id: str = marshaller.attrib(deserializer=str)

    #: The name of this account.
    #:
    #: :type: :obj:`str`
    name: str = marshaller.attrib(deserializer=str)


@marshaller.attrs(slots=True)
class PartialGuildIntegration(snowflakes.UniqueEntity, entities.Deserializable):
    """A partial representation of an integration, found in audit logs."""

    #: The name of this integration.
    #:
    #: :type: :obj:`str`
    name: str = marshaller.attrib(deserializer=str)

    #: The type of this integration.
    #:
    #: :type: :obj:`str`
    type: str = marshaller.attrib(deserializer=str)

    #: The account connected to this integration.
    #:
    #: :type: :obj:`IntegrationAccount`
    account: IntegrationAccount = marshaller.attrib(deserializer=IntegrationAccount.deserialize)


@marshaller.attrs(slots=True)
class GuildIntegration(snowflakes.UniqueEntity, entities.Deserializable):
    """Represents a guild integration object."""

    #: Whether this integration is enabled.
    #:
    #: :type: :obj:`bool`
    is_enabled: bool = marshaller.attrib(raw_name="enabled", deserializer=bool)

    #: Whether this integration is syncing subscribers/emojis.
    #:
    #: :type: :obj:`bool`
    is_syncing: bool = marshaller.attrib(raw_name="syncing", deserializer=bool)

    #: The ID of the managed role used for this integration's subscribers.
    #:
    #: :type: :obj:`snowflakes.Snowflake`
    role_id: typing.Optional[snowflakes.Snowflake] = marshaller.attrib(deserializer=snowflakes.Snowflake.deserialize)

    #: Whether users under this integration are allowed to use it's custom
    #: emojis.
    #:
    #: :type: :obj:`bool`, optional
    is_emojis_enabled: typing.Optional[bool] = marshaller.attrib(
        raw_name="enable_emoticons", deserializer=bool, if_undefined=None,
    )

    #: How members should be treated after their connected subscription expires
    #: This won't be enacted until after :attr:`expire_grace_period` passes.
    #:
    #: :type: :obj:`IntegrationExpireBehaviour`
    expire_behavior: IntegrationExpireBehaviour = marshaller.attrib(deserializer=IntegrationExpireBehaviour)

    #: The time delta for how many days users with expired subscriptions are
    #: given until :attr:`expire_behavior` is enacted out on them
    #:
    #: :type: :obj:`datetime.timedelta`
    expire_grace_period: datetime.timedelta = marshaller.attrib(
        deserializer=lambda delta: datetime.timedelta(days=delta),
    )

    #: The user this integration belongs to.
    #:
    #: :type: :obj:`users.User`
    user: users.User = marshaller.attrib(deserializer=users.User.deserialize)

    #: The datetime of when this integration's subscribers were last synced.
    #:
    #: :type: :obj:`datetime.datetime`
    last_synced_at: datetime.datetime = marshaller.attrib(
        raw_name="synced_at", deserializer=dates.parse_iso_8601_ts, if_none=None
    )


@marshaller.attrs(slots=True)
class GuildMemberBan(entities.HikariEntity, entities.Deserializable):
    """Used to represent guild bans."""

    #: The reason for this ban, will be :obj:`None` if no reason was given.
    #:
    #: :type: :obj:`str`, optional
    reason: str = marshaller.attrib(deserializer=str, if_none=None)

    #: The object of the user this ban targets.
    #:
    #: :type: :obj:`users.User`
    user: users.User = marshaller.attrib(deserializer=users.User.deserialize)


@marshaller.attrs(slots=True)
class UnavailableGuild(snowflakes.UniqueEntity, entities.Deserializable):
    """An unavailable guild object, received during gateway events such as
    the "Ready".
    An unavailable guild cannot be interacted with, and most information may
    be outdated if that is the case.
    """

    @property
    def is_unavailable(self) -> bool:
        """
        Whether this guild is unavailable or not, should always be :obj:`True`.
        """
        return True


@marshaller.attrs(slots=True)
class PartialGuild(snowflakes.UniqueEntity, entities.Deserializable):
    """This is a base object for any partial guild objects returned by the api
    where we are only given limited information."""

    #: The name of the guild.
    #:
    #: :type: :obj:`str`
    name: str = marshaller.attrib(deserializer=str)

    #: The hash for the guild icon, if there is one.
    #:
    #: :type: :obj:`str`, optional
    icon_hash: typing.Optional[str] = marshaller.attrib(raw_name="icon", deserializer=str, if_none=None)

    #: A set of the features in this guild.
    #:
    #: :type: :obj:`typing.Set` [ :obj:`GuildFeature` ]
    features: typing.Set[GuildFeature] = marshaller.attrib(
        deserializer=lambda features: {transformations.try_cast(f, GuildFeature, f) for f in features},
    )

    def format_icon_url(self, fmt: typing.Optional[str] = None, size: int = 2048) -> typing.Optional[str]:
        """Generate the url for this guild's custom icon, if set.

        Parameters
        ----------
        fmt : :obj:`str`
            The format to use for this url, defaults to ``png`` or ``gif``.
            Supports ``png``, ``jpeg``, `jpg`, ``webp`` and ``gif`` (when
            animated).
        size : :obj:`int`
            The size to set for the url, defaults to ``2048``.
            Can be any power of two between 16 and 2048.

        Returns
        -------
        :obj:`str`, optional
            The string url.
        """
        if self.icon_hash:
            # pylint: disable=E1101:
            if fmt is None and self.icon_hash.startswith("a_"):
                fmt = "gif"
            elif fmt is None:
                fmt = "png"
            return cdn.generate_cdn_url("icons", str(self.id), self.icon_hash, fmt=fmt, size=size)
        return None

    @property
    def icon_url(self) -> typing.Optional[str]:
        """The url for this guild's icon, if set."""
        return self.format_icon_url()


@marshaller.attrs(slots=True)
class Guild(PartialGuild):
    """A representation of a guild on Discord.

    Note
    ----
    If a guild object is considered to be unavailable, then the state of any
    other fields other than the :attr:`is_unavailable` and :attr:`id` members
    outdated, or incorrect. If a guild is unavailable, then the contents of any
    other fields should be ignored.
    """

    #: The hash of the splash for the guild, if there is one.
    #:
    #: :type: :obj:`str`, optional
    splash_hash: typing.Optional[str] = marshaller.attrib(raw_name="splash", deserializer=str, if_none=None)

    #: The hash of the discovery splash for the guild, if there is one.
    #:
    #: :type: :obj:`str`, optional
    discovery_splash_hash: typing.Optional[str] = marshaller.attrib(
        raw_name="discovery_splash", deserializer=str, if_none=None
    )

    #: The ID of the owner of this guild.
    #:
    #: :type: :obj:`snowflakes.Snowflake`
    owner_id: snowflakes.Snowflake = marshaller.attrib(deserializer=snowflakes.Snowflake.deserialize)

    #: The guild level permissions that apply to the bot user,
    #: Will be ``None`` when this object is retrieved through a REST request
    #: rather than from the gateway.
    #:
    #: :type: :obj:`_permissions.Permission`
    my_permissions: _permissions.Permission = marshaller.attrib(
        raw_name="permissions", deserializer=_permissions.Permission, if_undefined=None
    )

    #: The voice region for the guild.
    #:
    #: :type: :obj:`str`
    region: str = marshaller.attrib(deserializer=str)

    #: The ID for the channel that AFK voice users get sent to, if set for the
    #: guild.
    #:
    #: :type: :obj:`snowflakes.Snowflake`, optional
    afk_channel_id: typing.Optional[snowflakes.Snowflake] = marshaller.attrib(
        deserializer=snowflakes.Snowflake.deserialize, if_none=None
    )

    #: How long a voice user has to be AFK for before they are classed as being
    #: AFK and are moved to the AFK channel (:attr:`afk_channel_id`).
    #:
    #: :type: :obj:`datetime.timedelta`
    afk_timeout: datetime.timedelta = marshaller.attrib(
        deserializer=lambda seconds: datetime.timedelta(seconds=seconds)
    )

    # TODO: document when this is not specified.
    #: Defines if the guild embed is enabled or not. This information may not
    #: be present, in which case, it will be ``None`` instead.
    #:
    #: :type: :obj:`bool`, optional
    is_embed_enabled: typing.Optional[bool] = marshaller.attrib(
        raw_name="embed_enabled", if_undefined=False, deserializer=bool
    )

    #: The channel ID that the guild embed will generate an invite to, if
    #: enabled for this guild. Will be ``None`` if invites are disable for this
    #: guild's embed.
    #:
    #: :type: :obj:`snowflakes.Snowflake`, optional
    embed_channel_id: typing.Optional[snowflakes.Snowflake] = marshaller.attrib(
        deserializer=snowflakes.Snowflake.deserialize, if_none=None, if_undefined=None
    )

    #: The verification level required for a user to participate in this guild.
    #:
    #: :type: :obj:`GuildVerificationLevel`
    verification_level: GuildVerificationLevel = marshaller.attrib(deserializer=GuildVerificationLevel)

    #: The default setting for message notifications in this guild.
    #:
    #: :type: :obj:`GuildMessageNotificationsLevel`
    default_message_notifications: GuildMessageNotificationsLevel = marshaller.attrib(
        deserializer=GuildMessageNotificationsLevel
    )

    #: The setting for the explicit content filter in this guild.
    #:
    #: :type: :obj:`GuildExplicitContentFilterLevel`
    explicit_content_filter: GuildExplicitContentFilterLevel = marshaller.attrib(
        deserializer=GuildExplicitContentFilterLevel
    )

    #: The roles in this guild, represented as a mapping of ID to role object.
    #:
    #: :type: :obj:`typing.Mapping` [ :obj:`snowflakes.Snowflake`, :obj:`GuildRole` ]
    roles: typing.Mapping[snowflakes.Snowflake, GuildRole] = marshaller.attrib(
        deserializer=lambda roles: {r.id: r for r in map(GuildRole.deserialize, roles)},
    )

    #: The emojis that this guild provides, represented as a mapping of ID to
    #: emoji object.
    #:
    #: :type: :obj:`typing.Mapping` [ :obj:`snowflakes.Snowflake`, :obj:`_emojis.GuildEmoji` ]
    emojis: typing.Mapping[snowflakes.Snowflake, _emojis.GuildEmoji] = marshaller.attrib(
        deserializer=lambda emojis: {e.id: e for e in map(_emojis.GuildEmoji.deserialize, emojis)},
    )

    #: The required MFA level for users wishing to participate in this guild.
    #:
    #: :type: :obj:`GuildMFALevel`
    mfa_level: GuildMFALevel = marshaller.attrib(deserializer=GuildMFALevel)

    #: The ID of the application that created this guild, if it was created by
    #: a bot. If not, this is always ``None``.
    #:
    #: :type: :obj:`snowflakes.Snowflake`, optional
    application_id: typing.Optional[snowflakes.Snowflake] = marshaller.attrib(
        deserializer=snowflakes.Snowflake.deserialize, if_none=None
    )

    #: Whether the guild is unavailable or not.
    #:
    #: This information is only available if the guild was sent via a
    #: ``GUILD_CREATE`` event. If the guild is received from any other place,
    #: this will always be ``None``.
    #:
    #: An unavailable guild cannot be interacted with, and most information may
    #: be outdated if that is the case.
    is_unavailable: typing.Optional[bool] = marshaller.attrib(
        raw_name="unavailable", if_undefined=None, deserializer=bool
    )

    # TODO: document in which cases this information is not available.
    #: Describes whether the guild widget is enabled or not. If this information
    #: is not present, this will be ``None``.
    #:
    #: :type: :obj:`bool`, optional
    is_widget_enabled: typing.Optional[bool] = marshaller.attrib(
        raw_name="widget_enabled", if_undefined=None, deserializer=bool
    )

    #: The channel ID that the widget's generated invite will send the user to,
    #: if enabled. If this information is unavailable, this will be ``None``.
    #:
    #: :type: :obj:`snowflakes.Snowflake`, optional
    widget_channel_id: typing.Optional[snowflakes.Snowflake] = marshaller.attrib(
        if_undefined=None, if_none=None, deserializer=snowflakes.Snowflake.deserialize
    )

    #: The ID of the system channel (where welcome messages and Nitro boost
    #: messages are sent), or ``None`` if it is not enabled.
    #:
    #: :type: :obj:`snowflakes.Snowflake`, optional
    system_channel_id: typing.Optional[snowflakes.Snowflake] = marshaller.attrib(
        if_none=None, deserializer=snowflakes.Snowflake.deserialize
    )

    #: Flags for the guild system channel to describe which notification
    #: features are suppressed.
    #:
    #: :type: :obj:`GuildSystemChannelFlag`
    system_channel_flags: GuildSystemChannelFlag = marshaller.attrib(deserializer=GuildSystemChannelFlag)

    #: The ID of the channel where guilds with the :obj:`GuildFeature.PUBLIC`
    #: :attr:`features` display rules and guidelines. If the
    #: :obj:`GuildFeature.PUBLIC` feature is not defined, then this is ``None``.
    #:
    #: :type: :obj:`snowflakes.Snowflake`, optional
    rules_channel_id: typing.Optional[snowflakes.Snowflake] = marshaller.attrib(
        if_none=None, deserializer=snowflakes.Snowflake.deserialize
    )

    #: The date and time that the bot user joined this guild.
    #:
    #: This information is only available if the guild was sent via a
    #: ``GUILD_CREATE`` event. If the guild is received from any other place,
    #: this will always be ``None``.
    #:
    #: :type: :obj`datetime.datetime`, optional
    joined_at: typing.Optional[datetime.datetime] = marshaller.attrib(
        deserializer=dates.parse_iso_8601_ts, if_undefined=None
    )

    #: Whether the guild is considered to be large or not.
    #:
    #: This information is only available if the guild was sent via a
    #: ``GUILD_CREATE`` event. If the guild is received from any other place,
    #: this will always be ``None``.
    #:
    #: The implications of a large guild are that presence information will
    #: not be sent about members who are offline or invisible.
    #:
    #: :type: :obj:`bool`, optional
    is_large: typing.Optional[bool] = marshaller.attrib(raw_name="large", if_undefined=None, deserializer=bool)

    #: The number of members in this guild.
    #:
    #: This information is only available if the guild was sent via a
    #: ``GUILD_CREATE`` event. If the guild is received from any other place,
    #: this will always be ``None``.
    #:
    #: :type: :obj:`int`, optional
    member_count: typing.Optional[int] = marshaller.attrib(if_undefined=None, deserializer=int)

    #: A mapping of ID to the corresponding guild members in this guild.
    #:
    #: This information is only available if the guild was sent via a
    #: ``GUILD_CREATE`` event. If the guild is received from any other place,
    #: this will always be ``None``.
    #:
    #: Additionally, any offline members may not be included here, especially
    #: if there are more members than the large threshold set for the gateway
    #: this object was send with.
    #:
    #: This information will only be updated if your shards have the correct
    #: intents set for any update events.
    #:
    #: Essentially, you should not trust the information here to be a full
    #: representation. If you need complete accurate information, you should
    #: query the members using the appropriate API call instead.
    #:
    #: :type: :obj:`typing.Mapping` [ :obj:`snowflakes.Snowflake`, :obj:`GuildMember` ], optional
    members: typing.Optional[typing.Mapping[snowflakes.Snowflake, GuildMember]] = marshaller.attrib(
        deserializer=lambda members: {m.user.id: m for m in map(GuildMember.deserialize, members)}, if_undefined=None,
    )

    #: A mapping of ID to the corresponding guild channels in this guild.
    #:
    #: This information is only available if the guild was sent via a
    #: ``GUILD_CREATE`` event. If the guild is received from any other place,
    #: this will always be ``None``.
    #:
    #: Additionally, any channels that you lack permissions to see will not be
    #: defined here.
    #:
    #: This information will only be updated if your shards have the correct
    #: intents set for any update events.
    #:
    #: To retrieve a list of channels in any other case, you should make an
    #: appropriate API call to retrieve this information.
    #:
    #: :type: :obj:`typing.Mapping` [ :obj:`snowflakes.Snowflake`, :obj:`GuildChannel` ], optional
    channels: typing.Optional[typing.Mapping[snowflakes.Snowflake, GuildChannel]] = marshaller.attrib(
        deserializer=lambda guild_channels: {c.id: c for c in map(parse_guild_channel, guild_channels)},
        if_undefined=None,
    )

    #: A mapping of member ID to the corresponding presence information for
    #: the given member, if available.
    #:
    #: This information is only available if the guild was sent via a
    #: ``GUILD_CREATE`` event. If the guild is received from any other place,
    #: this will always be ``None``.
    #:
    #: Additionally, any channels that you lack permissions to see will not be
    #: defined here.
    #:
    #: This information will only be updated if your shards have the correct
    #: intents set for any update events.
    #:
    #: To retrieve a list of presences in any other case, you should make an
    #: appropriate API call to retrieve this information.
    #:
    #: :type: :obj:`typing.Mapping` [ :obj:`snowflakes.Snowflake`, :obj:`GuildMemberPresence` ], optional
    presences: typing.Optional[typing.Mapping[snowflakes.Snowflake, GuildMemberPresence]] = marshaller.attrib(
        deserializer=lambda presences: {p.user.id: p for p in map(GuildMemberPresence.deserialize, presences)},
        if_undefined=None,
    )

    #: The maximum number of presences for the guild. If this is ``None``, then
    #: the default value is used (currently 5000).
    #:
    #: :type: :obj:`int`, optional
    max_presences: typing.Optional[int] = marshaller.attrib(if_none=None, if_undefined=None, deserializer=int)

    #: The maximum number of members allowed in this guild.
    #:
    #: This information may not be present, in which case, it will be ``None``.
    #:
    #: :type: :obj:`int`, optional
    max_members: typing.Optional[int] = marshaller.attrib(if_undefined=None, deserializer=int)

    #: The vanity URL code for the guild's vanity URL.
    #: This is only present if :obj:`GuildFeatures.VANITY_URL` is in the
    #: :attr:`features` for this guild. If not, this will always be ``None``.
    #:
    #: :type: :obj:`str`, optional
    vanity_url_code: typing.Optional[str] = marshaller.attrib(if_none=None, deserializer=str)

    #: The guild's description.
    #:
    #: This is only present if certain :attr:`features` are set in this guild.
    #: Otherwise, this will always be ``None``. For all other purposes, it is
    #: ``None``.
    #:
    #: :type: :obj:`str`, optional
    description: typing.Optional[str] = marshaller.attrib(if_none=None, deserializer=str)

    #: The hash for the guild's banner.
    #: This is only present if the guild has :obj:`GuildFeatures.BANNER` in the
    #: :attr:`features` for this guild. For all other purposes, it is ``None``.
    #:
    #: :type: :obj:`str`, optional
    banner_hash: typing.Optional[str] = marshaller.attrib(raw_name="banner", if_none=None, deserializer=str)

    #: The premium tier for this guild.
    #:
    #: :type: :obj:`GuildPremiumTier`
    premium_tier: GuildPremiumTier = marshaller.attrib(deserializer=GuildPremiumTier)

    #: The number of nitro boosts that the server currently has. This
    #: information may not be present, in which case, it will be ``None``.
    #:
    #: :type: :obj:`int`, optional
    premium_subscription_count: typing.Optional[int] = marshaller.attrib(if_undefined=None, deserializer=int)

    #: The preferred locale to use for this guild.
    #:
    #: This can only be change if :obj:`GuildFeatures.PUBLIC` is in the
    #: :attr:`features` for this guild and will otherwise default to ``en-US```.
    #:
    #: :type: :obj:`str`
    preferred_locale: str = marshaller.attrib(deserializer=str)

    #: The channel ID of the channel where admins and moderators receive notices
    #: from Discord.
    #:
    #: This is only present if :obj:`GuildFeatures.PUBLIC` is in the
    #: :attr:`features` for this guild. For all other purposes, it should be
    #: considered to be ``None``.
    #:
    #: :type: :obj:`snowflakes.Snowflake`, optional
    public_updates_channel_id: typing.Optional[snowflakes.Snowflake] = marshaller.attrib(
        if_none=None, deserializer=snowflakes.Snowflake.deserialize
    )

    def format_splash_url(self, fmt: str = "png", size: int = 2048) -> typing.Optional[str]:
        """Generate the url for this guild's splash image, if set.

        Parameters
        ----------
        fmt : :obj:`str`
            The format to use for this url, defaults to ``png``.
            Supports ``png``, ``jpeg``, ``jpg`` and ``webp``.
        size : :obj:`int`
            The size to set for the url, defaults to ``2048``.
            Can be any power of two between 16 and 2048.

        Returns
        -------
        :obj:`str`, optional
            The string url.
        """
        if self.splash_hash:
            return cdn.generate_cdn_url("splashes", str(self.id), self.splash_hash, fmt=fmt, size=size)
        return None

    @property
    def splash_url(self) -> typing.Optional[str]:
        """The url for this guild's splash, if set."""
        return self.format_splash_url()

    def format_discovery_splash_url(self, fmt: str = "png", size: int = 2048) -> typing.Optional[str]:
        """Generate the url for this guild's discovery splash image, if set.

        Parameters
        ----------
        fmt : :obj:`str`
            The format to use for this url, defaults to ``png``.
            Supports ``png``, ``jpeg``, ``jpg`` and ``webp``.
        size : :obj:`int`
            The size to set for the url, defaults to ``2048``.
            Can be any power of two between 16 and 2048.

        Returns
        -------
        :obj:`str`, optional
            The string url.
        """
        if self.discovery_splash_hash:
            return cdn.generate_cdn_url(
                "discovery-splashes", str(self.id), self.discovery_splash_hash, fmt=fmt, size=size
            )
        return None

    @property
    def discovery_splash_url(self) -> typing.Optional[str]:
        """The url for this guild's discovery splash, if set."""
        return self.format_discovery_splash_url()

    def format_banner_url(self, fmt: str = "png", size: int = 2048) -> typing.Optional[str]:
        """Generate the url for this guild's banner image, if set.

        Parameters
        ----------
        fmt : :obj:`str`
            The format to use for this url, defaults to ``png``.
            Supports ``png``, ``jpeg``, ``jpg`` and ``webp``.
        size : :obj:`int`
            The size to set for the url, defaults to ``2048``.
            Can be any power of two between 16 and 2048.

        Returns
        -------
        :obj:`str`, optional
            The string url.
        """
        if self.banner_hash:
            return cdn.generate_cdn_url("banners", str(self.id), self.banner_hash, fmt=fmt, size=size)
        return None

    @property
    def banner_url(self) -> typing.Optional[str]:
        """The url for this guild's banner, if set."""
        return self.format_banner_url()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""An executable module to be used to test that the gateway works as intended.

This is only for use by developers of this library, regular users do not need
to use this.
"""
import logging


import click

from hikari.core.clients import gateway_client
from hikari.core.clients import gateway_config
from hikari.core.clients import protocol_config

logger_levels = ("DEBUG", "INFO", "WARNING", "ERROR", "NOTSET")


@click.command()
@click.option("--compression", default=True, type=click.BOOL, help="Enable or disable gateway compression.")
@click.option("--debug", default=False, type=click.BOOL, help="Enable or disable debug mode.")
@click.option("--logger", default="INFO", type=click.Choice(logger_levels), help="Logger verbosity.")
@click.option("--shards", default=1, type=click.IntRange(min=1), help="The number of shards to explicitly use.")
@click.option("--token", required=True, envvar="TOKEN", help="The token to use to authenticate with Discord.")
@click.option("--url", default="wss://gateway.discord.gg/", help="The websocket URL to connect to.")
@click.option("--verify-ssl", default=True, type=click.BOOL, help="Enable or disable SSL verification.")
@click.option("--version", default=7, type=click.IntRange(min=6), help="Version of the gateway to use.")
def run_gateway(compression, debug, logger, shards, token, url, verify_ssl, version):
    logging.basicConfig(level=logger)

    client = gateway_client.GatewayClient(
        config=gateway_config.GatewayConfig(
            debug=debug,
            protocol=protocol_config.HTTPProtocolConfig(verify_ssl=verify_ssl),
            shard_config=gateway_config.ShardConfig(shard_count=shards),
            token=token,
            use_compression=compression,
            version=version,
        ),
        url=url,
    )

    client.run()


run_gateway()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Assertions of things. 
These are functions that validate a value, expected to return the value on success but error
on any failure.
"""
__all__ = [
    "assert_that",
    "assert_not_none",
    "assert_in_range",
]

import typing

ValueT = typing.TypeVar("ValueT")
BaseTypeInstanceT = typing.TypeVar("BaseTypeInstanceT")


def assert_that(condition: bool, message: str = None, error_type: type = ValueError) -> None:
    """Raises a :obj:`ValueError` with the optional description if the given condition is falsified."""
    if not condition:
        raise error_type(message or "condition must not be False")


def assert_not_none(value: ValueT, message: typing.Optional[str] = None) -> ValueT:
    """Raises a :obj:`ValueError` with the optional description if the given value is ``None``."""
    if value is None:
        raise ValueError(message or "value must not be None")
    return value


def assert_in_range(value, min_inclusive, max_inclusive, name: str = None):
    """Raise a value error if a value is not in the range [min, max]"""
    if not (min_inclusive <= value <= max_inclusive):
        name = name or "The value"
        raise ValueError(f"{name} must be in the inclusive range of {min_inclusive} and {max_inclusive}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Gateway and sharding configuration objects and options."""
__all__ = ["GatewayConfig", "ShardConfig"]

import datetime
import re
import typing

from hikari.core import entities
from hikari.core import gateway_entities
from hikari.core.clients import protocol_config
from hikari.internal_utilities import assertions
from hikari.internal_utilities import dates
from hikari.internal_utilities import marshaller
from hikari.net import codes as net_codes


def _parse_shard_info(payload):
    range_matcher = re.search(r"(\d+)\s*(\.{2,3})\s*(\d+)", payload)

    if not range_matcher:
        if isinstance(payload, str):
            payload = int(payload)

        if isinstance(payload, int):
            return [payload]

        raise ValueError('expected shard_ids to be one of int, list of int, or range string ("x..y")')

    minimum, range_mod, maximum = range_matcher.groups()
    minimum, maximum = int(minimum), int(maximum)
    if len(range_mod) == 3:
        maximum += 1

    return [*range(minimum, maximum)]


@marshaller.attrs(kw_only=True, init=False)
class ShardConfig(entities.HikariEntity, entities.Deserializable):
    """Manual sharding configuration.

    All fields are optional kwargs that can be passed to the constructor.

    "Deserialized" and "unspecified" defaults are only applicable if you
    create the object using :meth:`deserialize`.
    """

    #: The shard IDs to produce shard connections for.
    #:
    #: If being deserialized, this can be several formats.
    #:     ``12``, ``"12"``:
    #:         A specific shard ID.
    #:     ``[0, 1, 2, 3, 8, 9, 10]``, ``["0", "1", "2", "3", "8", "9", "10"]``:
    #:         A sequence of shard IDs.
    #:     ``"5..16"``:
    #:         A range string. Two periods indicate a range of ``[5, 16)``
    #:         (inclusive beginning, exclusive end).
    #:     ``"5...16"``:
    #:         A range string. Three periods indicate a range of
    #:         ``[5, 17]`` (inclusive beginning, inclusive end).
    #:     ``None``:
    #:         The ``shard_count`` will be considered and that many shards will
    #:         be created for you.
    #:
    #: :type: :obj:`typing.Sequence` [ :obj:`int` ]
    shard_ids: typing.Sequence[int] = marshaller.attrib(
        deserializer=_parse_shard_info, if_none=None, if_undefined=None,
    )

    #: The number of shards the entire distributed application should consist
    #: of. If you run multiple instances of the bot, you should ensure this
    #: value is consistent.
    #:
    #: :type: :obj:`int`
    shard_count: int = marshaller.attrib(deserializer=int)

    def __init__(self, *, shard_ids: typing.Optional[typing.Iterable[int]] = None, shard_count: int) -> None:
        self.shard_ids = [*shard_ids] if shard_ids else [*range(shard_count)]

        for shard_id in self.shard_ids:
            assertions.assert_that(shard_id < shard_count, "shard_count must be greater than any shard ids")

        self.shard_count = shard_count


@marshaller.attrs(kw_only=True)
class GatewayConfig(entities.HikariEntity, entities.Deserializable):
    """Gateway and sharding configuration.

    All fields are optional kwargs that can be passed to the constructor.

    "Deserialized" and "unspecified" defaults are only applicable if you
    create the object using :meth:`deserialize`.
    """

    #: Whether to enable debugging mode for the generated shards. Usually you
    #: don't want to enable this.
    #:
    #: :type: :obj:`bool`
    debug: bool = marshaller.attrib(deserializer=bool, if_undefined=False, default=False)

    #: The initial activity to set all shards to when starting the gateway. If
    #: ``None``, then no activity will be set.
    #:
    #: :type: :obj:`GatewayActivity`, optional
    initial_activity: typing.Optional[gateway_entities.GatewayActivity] = marshaller.attrib(
        deserializer=gateway_entities.GatewayActivity.deserialize, if_none=None, if_undefined=None, default=None
    )

    # TODO: implement enum for this
    #: The initial status to set the shards to when starting the gateway.
    #:
    #: :type: :obj:`str`
    initial_status: str = marshaller.attrib(deserializer=str, if_undefined=lambda: "online", default="online")

    #: Whether to show up as AFK or not on sign-in.
    #:
    #: :type: :obj:`bool`
    initial_is_afk: bool = marshaller.attrib(deserializer=bool, if_undefined=False, default=False)

    #: The idle time to show on signing in, or ``None`` to not show an idle
    #: time.
    #:
    #: :type: :obj:`datetime.datetime`, optional
    initial_idle_since: typing.Optional[datetime.datetime] = marshaller.attrib(
        deserializer=dates.unix_epoch_to_ts, if_none=None, if_undefined=None, default=None
    )

    #: The intents to use for the connection.
    #:
    #: If being deserialized, this can be an integer bitfield, or a sequence of
    #: intent names. If
    #: unspecified, this will be set to ``None``.
    #:
    #: :type: :obj:`hikari.net.codes.GatewayIntent`, optional
    #:
    #: Examples
    #: --------
    #:
    #: .. code-block:: python
    #:
    #:    # Python example
    #:    GatewayIntent.GUILDS | GatewayIntent.GUILD_MESSAGES
    #:
    #: ..code-block:: js
    #:
    #:    // JSON example, using explicit bitfield values
    #:    513
    #:    // JSON example, using an array of names
    #:    [ "GUILDS", "GUILD_MESSAGES" ]
    #:
    #: See :obj:`hikari.net.codes.GatewayIntent` for valid names of
    #: intents you can use. Integer values are as documented on Discord's
    #: developer portal.
    #:
    #: Warnings
    #: --------
    #: If you are using the V7 gateway implementation, you will NEED to provide
    #: explicit intent values for this field in order to get online.
    #: Additionally, intents that are classed by Discord as being privileged
    #: will require you to whitelist your application in order to use them.
    #:
    #: If you are using the V6 gateway implementation, setting this to ``None``
    #: will simply opt you into every event you can subscribe to.
    intents: typing.Optional[net_codes.GatewayIntent] = marshaller.attrib(
        deserializer=lambda value: marshaller.dereference_int_flag(net_codes.GatewayIntent, value),
        if_undefined=None,
        default=None,
    )

    #: The large threshold to use.
    large_threshold: int = marshaller.attrib(deserializer=int, if_undefined=lambda: 250, default=True)

    #: Low level protocol details, such as proxy configuration and SSL settings.
    #:
    #: This is only used while creating the HTTP connection that the websocket
    #: is upgraded from.
    #:
    #: If unspecified, defaults are used.
    #:
    #: :type: :obj:`hikari.core.protocol_config.HTTPProtocolConfig`
    protocol: typing.Optional[protocol_config.HTTPProtocolConfig] = marshaller.attrib(
        deserializer=protocol_config.HTTPProtocolConfig.deserialize, if_undefined=None, default=None,
    )

    #: Manual sharding configuration to use. If this is ``None``, or
    #: unspecified, then auto sharding configuration will be performed for you
    #: based on defaults suggested by Discord.
    #:
    #: :type: :obj:`ShardConfig`, optional
    shard_config: typing.Optional[ShardConfig] = marshaller.attrib(
        deserializer=ShardConfig.deserialize, if_undefined=None, default=None,
    )

    #: The token to use, if applicable.
    #:
    #: If ``None`` or not specified, whatever is in the global token field on
    #: the config will be used. Note that you will have to specify this value
    #: somewhere; you will not be able to connect without it.
    #:
    #: :type: :obj:`str`, optional
    token: typing.Optional[str] = marshaller.attrib(deserializer=str, if_none=None, if_undefined=None, default=None)

    #: Whether to use zlib compression on the gateway for inbound messages or
    #: not. Usually you want this turned on.
    #:
    #: :type: :obj:`bool`
    use_compression: bool = marshaller.attrib(deserializer=bool, if_undefined=True, default=True)

    #: The gateway API version to use.
    #:
    #: If unspecified, then V6 is used.
    #:
    #: :type: :obj:`int`
    version: int = marshaller.attrib(deserializer=int, if_undefined=lambda: 6, default=6)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""The core API for interacting with Discord directly."""

# Do I need this? It still resolves without adding these in...?
from hikari.core import channels
from hikari.core import clients
from hikari.core import entities
from hikari.core import events
from hikari.core import gateway_entities
from hikari.core import guilds
from hikari.core import invites
from hikari.core import messages
from hikari.core import oauth2
from hikari.core import permissions
from hikari.core import snowflakes
from hikari.core import users
from hikari.core import webhooks

# Import everything into this namespace.
from hikari.core.channels import *
from hikari.core.clients import *
from hikari.core.colors import *
from hikari.core.colours import *
from hikari.core.embeds import *
from hikari.core.emojis import *
from hikari.core.entities import *
from hikari.core.events import *
from hikari.core.gateway_entities import *
from hikari.core.guilds import *
from hikari.core.invites import *
from hikari.core.messages import *
from hikari.core.oauth2 import *
from hikari.core.permissions import *
from hikari.core.snowflakes import *
from hikari.core.users import *
from hikari.core.voices import *
from hikari.core.webhooks import *

__all__ = [
    *channels.__all__,
    *clients.__all__,
    *entities.__all__,
    *events.__all__,
    *gateway_entities.__all__,
    *guilds.__all__,
    *invites.__all__,
    *messages.__all__,
    *oauth2.__all__,
    *permissions.__all__,
    *snowflakes.__all__,
    *users.__all__,
    *webhooks.__all__,
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Core application configuration objects and options."""
__all__ = ["AppConfig"]

import typing

from hikari.core import entities
from hikari.core.clients import http_config, gateway_config
from hikari.internal_utilities import marshaller


@marshaller.attrs(kw_only=True)
class AppConfig(entities.HikariEntity, entities.Deserializable):
    """Root application configuration object.

    All fields are optional kwargs that can be passed to the constructor.

    "Deserialized" and "unspecified" defaults are only applicable if you
    create the object using :meth:`deserialize`.

    Examples
    --------

    Initializing programatically:
        .. code-block:: python

            # Basic config
            config = AppConfig(token="1a2b3c4da9089288a.23rhagaa8==")

        .. code-block:: python

            # A more complicated config example
            config = AppConfig(
                gateway=GatewayConfig(
                    protocol=HTTPProtocolConfig(
                        allow_redirects=False,
                        proxy_auth=aiohttp.BasicAuth("username", "password"),
                        proxy_url="http://my.first.proxy.net:8080",
                        request_timeout=30.0,
                        verify_ssl=False,     # heresy! do not do this!
                    ),
                    sharding=ShardConfig(
                        shard_ids=range(0, 10),
                        shard_count=10
                    ),
                    version=6,
                ),
                http=HTTPConfig(
                    protocol=HTTPProtocolConfig(
                        allow_redirects=True,
                        proxy_auth=aiohttp.BasicAuth.decode("Basic dXNlcm5hbWU6cGFzc3dvcmQ="),
                        proxy_url="http://my.other.proxy.net:8080",
                        request_timeout=30.0,
                        ssl_context=mybot.utils.ssl.MySSLContext,
                        verify_ssl=True
                    ),
                    version=7,
                ),
                token="1a2b3c4da9089288a.23rhagaa8==",
            )

    Initializing from a file:
        .. code-block:: python

            # loading a JSON file
            with open("foo.json") as fp:
                config = AppConfig.deserialize(json.load(fp))

        .. code-block:: js

            /* basic config */
            {
                "token": "1a2b3c4da9089288a.23rhagaa8=="
            }

        .. code-block:: js

            /* a more complicated config example */
            {
                "gateway": {
                    "protocol": {
                        "allow_redirects": false,
                        "proxy_auth": "Basic dXNlcm5hbWU6cGFzc3dvcmQ=",
                        "proxy_url": "http://my.first.proxy.net:8080",
                        "request_timeout": 30.0,
                        "verify_ssl": false      // heresy, do not do this!
                    },
                    "sharding": {
                        "shard_ids": "0..10",
                        "shard_count": 10
                    },
                    "version": 6
                },
                "http": {
                    "protocol": {
                        "allow_redirects": true,
                        "proxy_auth": "Basic dXNlcm5hbWU6cGFzc3dvcmQ=",
                        "proxy_url": "http://my.other.proxy.net:8080",
                        "request_timeout": 30.0,
                        "ssl_context": "mybot.utils.ssl#MySSLContext",
                        "verify_ssl": true
                    },
                    "version": 7
                },
                "token": "1a2b3c4da9089288a.23rhagaa8=="
            }

    Of course, comments are not valid in actual standard JSON; I added them
    simply for reference for this example.
    """

    #: The HTTP configuration to use.
    #:
    #: If unspecified or None, then this will be a set of default values.
    #:
    #: :type: :obj:`hikari.core.http_config.HTTPConfig`, optional
    http: typing.Optional[http_config.HTTPConfig] = marshaller.attrib(
        deserializer=http_config.HTTPConfig.deserialize, if_none=None, if_undefined=None, default=None
    )

    #: The gateway configuration to use.
    #:
    #: If unspecified or None, then this will be a set of default values.
    #:
    #: :type: :obj:`hikari.core.gateway_config.GatewayConfig`, optional
    gateway: typing.Optional[gateway_config.GatewayConfig] = marshaller.attrib(
        deserializer=gateway_config.GatewayConfig.deserialize, if_none=None, if_undefined=None, default=None
    )

    #: The global token to use, if applicable. This can be overridden for each
    #: component that requires it.
    #:
    #: Note that this should not start with ``Bot`` or ``Bearer``. This is
    #: detected automatically.
    #:
    #: :type: :obj:`str`, optional
    token: typing.Optional[str] = marshaller.attrib(deserializer=str, if_none=None, if_undefined=None, default=None)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""IO utilities."""
__all__ = ["make_resource_seekable", "FileLikeT", "BytesLikeT"]

import io
import typing


def make_resource_seekable(resource: typing.Any) -> typing.Union[io.BytesIO, io.StringIO]:
    """Given some representation of data, make a seekable resource to use. 
    
    This supports :obj:`bytes`, :obj:`bytearray`, :obj:`memoryview`, and :obj:`str`. 
    Anything else is just returned.

    Parameters
    ----------
    resource : :obj:`typing.Any`
        The resource to check.

    Returns
    -------
    :obj:`typing.Union` [ :obj:`io.BytesIO`, :obj:`io.StringIO` ]
        An stream-compatible resource where possible.
    """
    if isinstance(resource, (bytes, bytearray)):
        resource = io.BytesIO(resource)
    elif isinstance(resource, memoryview):
        resource = io.BytesIO(resource.tobytes())
    elif isinstance(resource, str):
        resource = io.StringIO(resource)

    return resource


#: A bytes-like object, such as a :obj:`str`, raw :obj:`bytes`, or view across a bytes-like object.
BytesLikeT = typing.Union[bytes, bytearray, memoryview, str, io.StringIO, io.BytesIO]

#: Type description for any object that can be considered to be file-like.
FileLikeT = typing.Union[BytesLikeT, io.BufferedRandom, io.BufferedReader, io.BufferedRWPair]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe both DMs and guild
channels on Discord.
"""

__all__ = [
    "Channel",
    "ChannelType",
    "DMChannel",
    "PartialChannel",
    "PermissionOverwrite",
    "PermissionOverwriteType",
    "GroupDMChannel",
]

import enum
import typing

from hikari.core import entities
from hikari.core import snowflakes
from hikari.core import permissions
from hikari.internal_utilities import marshaller


@enum.unique
class ChannelType(enum.IntEnum):
    """The known channel types that are exposed to us by the api."""

    #: A text channel in a guild.
    GUILD_TEXT = 0
    #: A direct channel between two users.
    DM = 1
    #: A voice channel in a guild.
    GUILD_VOICE = 2
    #: A direct channel between multiple users.
    GROUP_DM = 3
    #: An category used for organizing channels in a guild.
    GUILD_CATEGORY = 4
    #: A channel that can be followed and can crosspost.
    GUILD_NEWS = 5
    #: A channel that show's a game's store page.
    GUILD_STORE = 6


@marshaller.attrs(slots=True)
class PartialChannel(snowflakes.UniqueEntity, entities.Deserializable):
    """Represents a channel where we've only received it's basic information,
    commonly received in rest responses.
    """

    #: This channel's name.
    #:
    #: :type: :obj:`str`
    name: str = marshaller.attrib(deserializer=str)

    #: This channel's type.
    #:
    #: :type: :obj:`ChannelType`
    type: ChannelType = marshaller.attrib(deserializer=ChannelType)


@enum.unique
class PermissionOverwriteType(str, enum.Enum):
    """The type of entity a Permission Overwrite targets."""

    #: A permission overwrite that targets all the members with a specific
    #: guild role.
    ROLE = "role"
    #: A permission overwrite that targets a specific guild member.
    MEMBER = "member"


@marshaller.attrs(slots=True)
class PermissionOverwrite(snowflakes.UniqueEntity, entities.Deserializable, entities.Serializable):
    """Represents permission overwrites for a channel or role in a channel."""

    #: The type of entity this overwrite targets.
    #:
    #: :type: :obj:`PermissionOverwriteType`
    type: PermissionOverwriteType = marshaller.attrib(deserializer=PermissionOverwriteType)

    #: The permissions this overwrite allows.
    #:
    #: :type: :obj:`permissions.Permission`
    allow: permissions.Permission = marshaller.attrib(deserializer=permissions.Permission)

    #: The permissions this overwrite denies.
    #:
    #: :type: :obj:`permissions.Permission`
    deny: permissions.Permission = marshaller.attrib(deserializer=permissions.Permission)

    @property
    def unset(self) -> permissions.Permission:
        return typing.cast(permissions.Permission, (self.allow | self.deny))


@marshaller.attrs(slots=True)
class Channel(PartialChannel):
    ...


@marshaller.attrs(slots=True)
class DMChannel(Channel):
    ...


@marshaller.attrs(slots=True)
class GroupDMChannel(DMChannel):
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Custom data structures and constant values."""
__all__ = [
    "EMPTY_SEQUENCE",
    "EMPTY_SET",
    "EMPTY_COLLECTION",
    "EMPTY_DICT",
    "EMPTY_GENERATOR_EXPRESSION",
    "WeakKeyDictionary",
]

import types
import typing

#: An immutable indexable container of elements with zero size.
import weakref

EMPTY_SEQUENCE: typing.Sequence = tuple()
#: An immutable unordered container of elements with zero size.
EMPTY_SET: typing.AbstractSet = frozenset()
#: An immutable container of elements with zero size.
EMPTY_COLLECTION: typing.Collection = tuple()
#: An immutable ordered mapping of key elements to value elements with zero size.
EMPTY_DICT: typing.Mapping = types.MappingProxyType({})
#: An empty generator expression that can be used as a placeholder, but never
#: yields anything.
EMPTY_GENERATOR_EXPRESSION = (_ for _ in EMPTY_COLLECTION)


K = typing.TypeVar("K")
V = typing.TypeVar("V")


class WeakKeyDictionary(weakref.WeakKeyDictionary, typing.MutableMapping[K, V]):
    """A dictionary that has weak references to the keys.

    This is a type-safe version of :obj:`weakref.WeakKeyDictionary`.
    """


class WeakValueDictionary(weakref.WeakValueDictionary, typing.MutableMapping[K, V]):
    """A dictionary that has weak references to the values.

    This is a type-safe version of :obj:`weakref.WeakValueDictionary`.
    """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Core errors that may be raised by this API implementation."""
__all__ = ["HikariError"]


class HikariError(RuntimeError):
    """Base for an error raised by this API.

    Any errors should derive from this.

    Note
    ----
    You should never initialize this exception directly.
    """

    __slots__ = ()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
__all__ = ["GatewayClient"]

import asyncio
import inspect
import time
import typing

from hikari.core import dispatcher
from hikari.core import events
from hikari.core.clients import gateway_config
from hikari.core.clients import shard_client
from hikari.core import state
from hikari.internal_utilities import loggers
from hikari.net import shard

ShardT = typing.TypeVar("ShardT", bound=shard_client.ShardClient)


class GatewayClient(typing.Generic[ShardT], shard_client.WebsocketClientBase, dispatcher.EventDispatcher):
    def __init__(
        self,
        config: gateway_config.GatewayConfig,
        url: str,
        *,
        dispatcher_impl: typing.Optional[dispatcher.EventDispatcher] = None,
        shard_type: typing.Type[ShardT] = shard_client.ShardClient,
    ) -> None:
        self.logger = loggers.get_named_logger(self)
        self.config = config
        self.event_dispatcher = dispatcher_impl if dispatcher_impl is not None else dispatcher.EventDispatcherImpl()
        self._websocket_event_types = self._websocket_events()
        self.state = state.StatefulStateManagerImpl()
        self._is_running = False
        self.shards: typing.Dict[int, ShardT] = {
            shard_id: shard_type(shard_id, config, self._handle_websocket_event_later, url)
            for shard_id in config.shard_config.shard_ids
        }

    async def start(self) -> None:
        """Start all shards.

        This safely starts all shards at the correct rate to prevent invalid
        session spam. This involves starting each shard sequentially with a
        5 second pause between each.
        """
        self._is_running = True
        self.logger.info("starting %s shard(s)", len(self.shards))
        start_time = time.perf_counter()
        for i, shard_id in enumerate(self.config.shard_config.shard_ids):
            if i > 0:
                await asyncio.sleep(5)

            shard_obj = self.shards[shard_id]
            await shard_obj.start()
        finish_time = time.perf_counter()

        self.logger.info("started %s shard(s) in approx %.2fs", len(self.shards), finish_time - start_time)

    async def join(self) -> None:
        await asyncio.gather(*(shard_obj.join() for shard_obj in self.shards.values()))

    async def close(self, wait: bool = True) -> None:
        if self._is_running:
            self.logger.info("stopping %s shard(s)", len(self.shards))
            start_time = time.perf_counter()
            try:
                await asyncio.gather(*(shard_obj.close(wait) for shard_obj in self.shards.values()))
            finally:
                finish_time = time.perf_counter()
                self.logger.info("stopped %s shard(s) in approx %.2fs", len(self.shards), finish_time - start_time)
                self._is_running = False

    async def wait_for(
        self,
        event_type: typing.Type[dispatcher.EventT],
        *,
        predicate: dispatcher.PredicateT,
        timeout: typing.Optional[float],
    ) -> dispatcher.EventT:
        """Wait for the given event type to occur.

        Parameters
        ----------
        event_type : :obj:`typing.Type` [ :obj:`events.HikariEvent` ]
            The name of the event to wait for.
        timeout : :obj:`float`, optional
            The timeout to wait for before cancelling and raising an
            :obj:`asyncio.TimeoutError` instead. If this is `None`, this will
            wait forever. Care must be taken if you use `None` as this may
            leak memory if you do this from an event listener that gets
            repeatedly called. If you want to do this, you should consider
            using an event listener instead of this function.
        predicate : ``def predicate(event) -> bool`` or ``async def predicate(event) -> bool``
            A function that takes the arguments for the event and returns True
            if it is a match, or False if it should be ignored.
            This can be a coroutine function that returns a boolean, or a
            regular function.

        Returns
        -------
        :obj:`asyncio.Future`:
            A future to await. When the given event is matched, this will be
            completed with the corresponding event body.

            If the predicate throws an exception, or the timeout is reached,
            then this will be set as an exception on the returned future.
        """
        return await self.event_dispatcher.wait_for(event_type, predicate=predicate, timeout=timeout)

    def _handle_websocket_event_later(self, conn: shard.ShardConnection, event_name: str, payload: typing.Any) -> None:
        # Run this asynchronously so that we can allow awaiting stuff like state management.
        asyncio.get_event_loop().create_task(self._handle_websocket_event(conn, event_name, payload))

    async def _handle_websocket_event(self, _: shard.ShardConnection, event_name: str, payload: typing.Any) -> None:
        try:
            event_type = self._websocket_event_types[event_name]
        except KeyError:
            pass
        else:
            event_payload = event_type.deserialize(payload)
            await self.state.on_event(event_payload)
            await self.event_dispatcher.dispatch_event(event_payload)

    def _websocket_events(self):
        # Look for anything that has the ___raw_ws_event_name___ class attribute
        # to each corresponding class where appropriate to do so. This provides
        # a quick and dirty event lookup mechanism that can be extended quickly
        # and has O(k) lookup time.

        types = {}

        def predicate(member):
            return inspect.isclass(member) and hasattr(member, "___raw_ws_event_name___")

        for name, cls in inspect.getmembers(events, predicate):
            raw_name = cls.___raw_ws_event_name___
            types[raw_name] = cls
            self.logger.debug("detected %s as a web socket event to listen for", name)

        self.logger.debug("detected %s web socket events to register from %s", len(types), events.__name__)

        return types

    def add_listener(
        self, event_type: typing.Type[dispatcher.EventT], callback: dispatcher.EventCallbackT
    ) -> dispatcher.EventCallbackT:
        return self.event_dispatcher.add_listener(event_type, callback)

    def remove_listener(
        self, event_type: typing.Type[dispatcher.EventT], callback: dispatcher.EventCallbackT
    ) -> dispatcher.EventCallbackT:
        return self.event_dispatcher.remove_listener(event_type, callback)

    def dispatch_event(self, event: events.HikariEvent) -> ...:
        return self.event_dispatcher.dispatch_event(event)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
__all__ = ["ShardState", "ShardClient"]

import abc
import asyncio
import contextlib
import datetime
import enum
import logging
import signal
import time
import typing

import aiohttp

from hikari.core import events
from hikari.core.clients import gateway_config
from hikari.core import gateway_entities
from hikari.internal_utilities import aio
from hikari.internal_utilities import loggers
from hikari.net import codes
from hikari.net import errors
from hikari.net import ratelimits
from hikari.net import shard

_EventT = typing.TypeVar("_EventT", bound=events.HikariEvent)


@enum.unique
class ShardState(enum.IntEnum):
    NOT_RUNNING = 0
    INITIALIZING = enum.auto()
    WAITING_FOR_READY = enum.auto()
    READY = enum.auto()
    RECONNECTING = enum.auto()
    STOPPING = enum.auto()
    STOPPED = enum.auto()


class WebsocketClientBase(abc.ABC):
    """Base for any socket-based communication medium to provide functionality
    for more automated control given certain method constraints.
    """

    logger: logging.Logger

    @abc.abstractmethod
    async def start(self):
        ...

    @abc.abstractmethod
    async def close(self, wait: bool = True):
        ...

    @abc.abstractmethod
    async def join(self):
        ...

    def run(self):
        loop = asyncio.get_event_loop()

        def sigterm_handler(*_):
            raise KeyboardInterrupt()

        ex = None

        try:
            with contextlib.suppress(NotImplementedError):
                # Not implemented on Windows
                loop.add_signal_handler(signal.SIGTERM, sigterm_handler)

            loop.run_until_complete(self.start())
            loop.run_until_complete(self.join())

            self.logger.info("client has shut down")

        except KeyboardInterrupt as _ex:
            self.logger.info("received signal to shut down client")
            loop.run_until_complete(self.close())
            # Apparently you have to alias except clauses or you get an
            # UnboundLocalError.
            ex = _ex
        finally:
            loop.run_until_complete(self.close(True))
            with contextlib.suppress(NotImplementedError):
                # Not implemented on Windows
                loop.remove_signal_handler(signal.SIGTERM)

        if ex:
            raise ex from ex


class ShardClient(WebsocketClientBase):
    """The primary interface for a single shard connection. This contains
    several abstractions to enable usage of the low level gateway network
    interface with the higher level constructs in :mod:`hikari.core`.

    Parameters
    ----------
    shard_id : :obj:`int`
        The ID of this specific shard.
    config : :obj:`gateway_config.GatewayConfig`
        The gateway configuration to use to initialize this shard.
    low_level_dispatch : :obj:`typing.Callable` [ [ :obj:`Shard`, :obj:`str`, :obj:`typing.Any` ] ]
        A function that is fed any low-level event payloads. This will consist
        of three arguments: an :obj:`Shard` which is this shard instance,
        a :obj:`str` of the raw event name, and any naive raw payload that was
        passed with the event. The expectation is the function passed here
        will pass the payload onto any event handling and state handling system
        to be transformed into a higher-level representation.
    url : :obj:`str`
        The URL to connect the gateway to.

    Notes
    -----
    Generally, you want to use :class:`GatewayClient` rather than this class
    directly, as that will handle sharding where enabled and applicable, and
    provides a few more bits and pieces that may be useful such as state
    management and event dispatcher integration. and If you want to customize
    this, you can subclass it and simply override anything you want.
    """

    __slots__ = (
        "logger",
        "_dispatch",
        "_client",
        "_status",
        "_activity",
        "_idle_since",
        "_is_afk",
        "_task",
        "_shard_state",
    )

    def __init__(
        self,
        shard_id: int,
        config: gateway_config.GatewayConfig,
        low_level_dispatch: typing.Callable[["ShardClient", str, typing.Any], None],
        url: str,
    ) -> None:
        self.logger = loggers.get_named_logger(self, shard_id)
        self._dispatch = low_level_dispatch
        self._activity = config.initial_activity
        self._idle_since = config.initial_idle_since
        self._is_afk = config.initial_is_afk
        self._status = config.initial_status
        self._shard_state = ShardState.NOT_RUNNING
        self._task = None
        self._client = shard.ShardConnection(
            compression=config.use_compression,
            connector=config.protocol.connector if config.protocol is not None else None,
            debug=config.debug,
            dispatch=self._dispatch,
            initial_presence=self._create_presence_pl(
                status=config.initial_status,
                activity=config.initial_activity,
                idle_since=config.initial_idle_since,
                is_afk=config.initial_is_afk,
            ),
            intents=config.intents,
            large_threshold=config.large_threshold,
            proxy_auth=config.protocol.proxy_auth if config.protocol is not None else None,
            proxy_headers=config.protocol.proxy_headers if config.protocol is not None else None,
            proxy_url=config.protocol.proxy_url if config.protocol is not None else None,
            session_id=None,
            seq=None,
            shard_id=shard_id,
            shard_count=config.shard_config.shard_count,
            ssl_context=config.protocol.ssl_context if config.protocol is not None else None,
            token=config.token,
            url=url,
            verify_ssl=config.protocol.verify_ssl if config.protocol is not None else None,
            version=config.version,
        )

    @property
    def client(self) -> shard.ShardConnection:
        """
        Returns
        -------
        :obj:`hikari.net.gateway.GatewayClient`
            The low-level gateway client used for this shard.
        """
        return self._client

    #: TODO: use enum
    @property
    def status(self) -> str:
        """
        Returns
        -------
        :obj:`str`
            The current user status for this shard.
        """
        return self._status

    @property
    def activity(self) -> typing.Optional[gateway_entities.GatewayActivity]:
        """
        Returns
        -------
        :obj:`hikari.core.gateway_entities.GatewayActivity`, optional
            The current activity for the user on this shard, or ``None`` if
            there is no activity.
        """
        return self._activity

    @property
    def idle_since(self) -> typing.Optional[datetime.datetime]:
        """
        Returns
        -------
        :obj:`datetime.datetime`, optional
            The timestamp when the user of this shard appeared to be idle, or
            ``None`` if not applicable.
        """
        return self._idle_since

    @property
    def is_afk(self) -> bool:
        """
        Returns
        -------
        :obj:`bool`
            ``True`` if the user is AFK, ``False`` otherwise.
        """
        return self._is_afk

    async def start(self):
        """Connect to the gateway on this shard and schedule tasks to keep this
        connection alive. Wait for the shard to dispatch a ``READY`` event, and
        then return.
        """
        if self._shard_state not in (ShardState.NOT_RUNNING, ShardState.STOPPED):
            raise RuntimeError("Cannot start a shard twice")

        self.logger.debug("starting shard")
        self._shard_state = ShardState.INITIALIZING
        self._task = asyncio.create_task(self._keep_alive())
        self.logger.info("waiting for READY")
        completed, _ = await asyncio.wait(
            [self._task, self._client.identify_event.wait()], return_when=asyncio.FIRST_COMPLETED
        )

        if self._task in completed:
            raise self._task.exception()

        self._shard_state = ShardState.WAITING_FOR_READY

        completed, _ = await asyncio.wait(
            [self._task, self._client.ready_event.wait()], return_when=asyncio.FIRST_COMPLETED
        )

        if self._task in completed:
            raise self._task.exception()

        self.logger.info("now READY")
        self._shard_state = ShardState.READY

    async def join(self) -> None:
        """Wait for the shard to shut down fully."""
        await self._task if self._task is not None else aio.completed_future()

    async def close(self, wait: bool = True) -> None:
        """Request that the shard shuts down.

        Parameters
        ----------
        wait : :obj:`bool`
            If ``True`` (default), then wait for the client to shut down fully.
            If ``False``, only send the signal to shut down, but do not wait
            for it explicitly.
        """
        if self._shard_state != ShardState.STOPPING:
            self._shard_state = ShardState.STOPPING
            self.logger.debug("stopping shard")
            await self._client.close()
            if wait:
                await self._task
            with contextlib.suppress():
                self._task.result()

    async def _keep_alive(self):
        back_off = ratelimits.ExponentialBackOff(maximum=None)
        last_start = time.perf_counter()
        do_not_back_off = True

        while True:
            try:
                if not do_not_back_off and time.perf_counter() - last_start < 30:
                    next_backoff = next(back_off)
                    self.logger.info(
                        "restarted within 30 seconds, will backoff for %ss", next_backoff,
                    )
                    await asyncio.sleep(next_backoff)
                else:
                    back_off.reset()

                last_start = time.perf_counter()

                do_not_back_off = False
                await self._client.connect()
                self.logger.critical("shut down silently! this shouldn't happen!")

            except aiohttp.ClientConnectorError as ex:
                self.logger.exception(
                    "failed to connect to Discord to initialize a websocket connection", exc_info=ex,
                )
            except errors.GatewayZombiedError:
                self.logger.warning("entered a zombie state and will be restarted")
            except errors.GatewayInvalidSessionError as ex:
                if ex.can_resume:
                    self.logger.warning("invalid session, so will attempt to resume")
                else:
                    self.logger.warning("invalid session, so will attempt to reconnect")

                if not ex.can_resume:
                    self._client.seq = None
                    self._client.session_id = None
                do_not_back_off = True
                await asyncio.sleep(5)
            except errors.GatewayMustReconnectError:
                self.logger.warning("instructed by Discord to reconnect")
                self._client.seq = None
                self._client.session_id = None
                do_not_back_off = True
                await asyncio.sleep(5)
            except errors.GatewayServerClosedConnectionError as ex:
                if ex.close_code in (
                    codes.GatewayCloseCode.RATE_LIMITED,
                    codes.GatewayCloseCode.SESSION_TIMEOUT,
                    codes.GatewayCloseCode.INVALID_SEQ,
                    codes.GatewayCloseCode.UNKNOWN_ERROR,
                    codes.GatewayCloseCode.SESSION_TIMEOUT,
                    codes.GatewayCloseCode.NORMAL_CLOSURE,
                ):
                    self.logger.warning("disconnected by Discord, will attempt to reconnect")
                else:
                    self.logger.error("disconnected by Discord, %s: %s", type(ex).__name__, ex.reason)
                    raise ex
            except errors.GatewayClientClosedError:
                self.logger.warning("shutting down")
                return
            except Exception as ex:
                self.logger.debug("propagating unexpected exception %s", exc_info=ex)
                raise ex

    async def update_presence(
        self,
        status: str = ...,  # TODO: use enum for status
        activity: typing.Optional[gateway_entities.GatewayActivity] = ...,
        idle_since: typing.Optional[datetime.datetime] = ...,
        is_afk: bool = ...,
    ) -> None:
        """Update the presence of the user for the shard.

        This will only update arguments that you explicitly specify a value for.
        Any arguments that you do not explicitly provide some value for will
        not be changed.

        Warnings
        --------
        This will fail if the shard is not online.

        Parameters
        ----------
        status : :obj:`str`
            The new status to set.
        activity : :obj:`hikari.core.gateway_entities.GatewayActivity`, optional
            The new activity to set.
        idle_since : :obj:`datetime.datetime`, optional
            The time to show up as being idle since, or ``None`` if not
            applicable.
        is_afk : :obj:`bool`
            ``True`` if the user should be marked as AFK, or ``False``
            otherwise.
        """
        status = self._status if status is ... else status
        activity = self._activity if activity is ... else activity
        idle_since = self._idle_since if idle_since is ... else idle_since
        is_afk = self._is_afk if is_afk is ... else is_afk

        presence = self._create_presence_pl(status=status, activity=activity, idle_since=idle_since, is_afk=is_afk)
        await self._client.update_presence(presence)

        # If we get this far, the update succeeded probably, or the gateway just died. Whatever.
        self._status = status
        self._activity = activity
        self._idle_since = idle_since
        self._is_afk = is_afk

    @staticmethod
    def _create_presence_pl(
        status: str,  # TODO: use enum for status
        activity: typing.Optional[gateway_entities.GatewayActivity],
        idle_since: typing.Optional[datetime.datetime],
        is_afk: bool,
    ) -> typing.Dict[str, typing.Any]:
        return {
            "status": status,
            "idle_since": idle_since.timestamp() if idle_since is not None else None,
            "game": activity.serialize() if activity is not None else None,
            "afk": is_afk,
        }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Provides mechanisms to cache results of calls lazily."""
__all__ = [
    "CachedFunctionT",
    "CachedPropertyFunctionT",
    "CachedFunction",
    "CachedProperty",
    "AsyncCachedProperty",
    "cached_function",
    "cached_property",
]

import asyncio
import functools
import inspect
import os
import typing

ReturnT = typing.TypeVar("ReturnT")
ClassT = typing.TypeVar("ClassT")
CallT = typing.Callable[..., ReturnT]
CachedFunctionT = typing.Callable[..., ReturnT]
CachedPropertyFunctionT = typing.Callable[[ClassT], ReturnT]

# Hacky workaround to Sphinx being unable to document cached properties. We simply make the
# decorators return their inputs when this is True.
__is_sphinx = os.getenv("SPHINX_IS_GENERATING_DOCUMENTATION") is not None


def __noop_decorator(func):  # pragma: no cover
    return func


class CachedFunction:
    """Wraps a call, some arguments, and some keyword arguments in a partial and stores the
    result of the call for later invocations.

    Warning
    -------
    This is not thread safe!
    """

    _sentinel = object()
    __slots__ = (
        "_call",
        "_value",
        "__qualname__",  # pylint: disable=class-variable-slots-conflict
        "__dict__",
        "__name__",
    )

    def __init__(self, call, args, kwargs):
        self._value = self._sentinel
        self.__qualname__ = getattr(call, "__qualname__", None)
        self.__name__ = getattr(call, "__name__", None)
        self.__dict__ = getattr(call, "__dict__", None)
        is_coro = inspect.iscoroutinefunction(call)
        call_wrapper = self._coroutine_fn_wrapper if is_coro else self._fn_wrapper
        self._call = call_wrapper(call, args, kwargs)

    def __call__(self) -> ReturnT:
        if self._value is self._sentinel:
            self._call()
        return self._value

    def _coroutine_fn_wrapper(self, call, args, kwargs):
        def fn_wrapper():
            self._value = asyncio.create_task(call(*args, **kwargs), name="pending CachedFunction coroutine completion")

        return fn_wrapper

    def _fn_wrapper(self, call, args, kwargs):
        def fn_wrapper():
            self._value = call(*args, **kwargs)

        return fn_wrapper


class CachedProperty:
    """A get/delete descriptor to wrap a no-args method which can cache the result of the
    call for future retrieval. Calling :func:`del` on the property will flush the cache.

    This will misbehave on class methods and static methods, and will not work on
    non-instance functions. For general functions, you should consider :obj:`CachedFunction`
    instead.
    """

    __slots__ = (
        "func",
        "_cache_attr",
        "__dict__",
        "__name__",
        "__qualname__",  # pylint: disable=class-variable-slots-conflict
    )

    def __init__(self, func: CachedPropertyFunctionT, cache_attr: typing.Optional[str]) -> None:
        self.func = func
        self._cache_attr = cache_attr or "_cp_" + func.__name__
        self.__dict__ = getattr(self.func, "__dict__", None)

    def __get__(self, instance: typing.Optional[ClassT], owner: typing.Type[ClassT]) -> ReturnT:
        if instance is None:
            return typing.cast(ReturnT, self)
        if not hasattr(instance, self._cache_attr):
            setattr(instance, self._cache_attr, self.func(instance))
        return getattr(instance, self._cache_attr)

    def __delete__(self, instance: ClassT):
        try:
            delattr(instance, self._cache_attr)
        except AttributeError:
            pass


class AsyncCachedProperty(CachedProperty):
    """Cached property implementation that supports awaitable coroutines."""

    __slots__ = ()

    def __get__(self, instance: typing.Optional[ClassT], owner: typing.Type[ClassT]) -> typing.Awaitable[ReturnT]:
        if instance is None:
            return typing.cast(ReturnT, self)

        if not hasattr(instance, self._cache_attr):
            setattr(
                instance,
                self._cache_attr,
                asyncio.create_task(self.func(instance), name="pending AsyncCachedProperty coroutine completion"),
            )
        return getattr(instance, self._cache_attr)


def cached_function(*args, **kwargs) -> typing.Callable[[CachedFunctionT], typing.Callable[[], ReturnT]]:
    """Create a wrapped cached call decorator. 
    
    This remembers the last result of the given call forever until cleared.

    Parameters
    -----------
    *args
        Any arguments to call the call with.
    **kwargs
        Any kwargs to call the call with.

    Note
    ----
    This is not useful for instance methods on classes, you should use
    a :obj:`CachedProperty` instead for those. You should also not expect
    thread safety here. Coroutines will be detected and dealt with as futures.
    This is lazily evaluated.
    """

    def decorator(func):
        return functools.wraps(func)(CachedFunction(func, args, kwargs))

    return decorator if not __is_sphinx else __noop_decorator


def cached_property(
    *, cache_name=None
) -> typing.Callable[[CachedPropertyFunctionT], typing.Union[CachedProperty, AsyncCachedProperty]]:
    """Makes a slots-compatible cached property.

    If using slots, you should specify the ``cache_name`` directly.
    """

    def decorator(func: CachedPropertyFunctionT) -> typing.Union[CachedProperty, AsyncCachedProperty]:
        cls = AsyncCachedProperty if asyncio.iscoroutinefunction(func) else CachedProperty
        return typing.cast(
            typing.Union[CachedProperty, AsyncCachedProperty], functools.wraps(func)(cls(func, cache_name))
        )

    return decorator if not __is_sphinx else __noop_decorator



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Asyncio extensions and utilities."""
__all__ = [
    "CoroutineFunctionT",
    "PartialCoroutineProtocolT",
    "completed_future",
]

import asyncio
import typing

ReturnT = typing.TypeVar("ReturnT", covariant=True)
CoroutineFunctionT = typing.Callable[..., typing.Coroutine[typing.Any, typing.Any, ReturnT]]


class PartialCoroutineProtocolT(typing.Protocol[ReturnT]):
    """Represents the type of a :obj:`functools.partial` wrapping an :mod:`asyncio` coroutine."""

    def __call__(self, *args, **kwargs) -> typing.Coroutine[None, None, ReturnT]:
        ...

    def __await__(self):
        ...


def completed_future(result: typing.Any = None) -> asyncio.Future:
    """Create a future on the current running loop that is completed, then return it.

    Parameters
    ---------
    result : :obj:`typing.Any`
        The value to set for the result of the future.

    Returns
    -------
    :obj:`asyncio.Future`
        The completed future.
    """
    future = asyncio.get_event_loop().create_future()
    future.set_result(result)
    return future



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Various utilities used internally within this API. 

These are not bound to the versioning contact, and are considered to be 
implementation detail that could change at any time, so should not be 
used outside this library.
"""
from hikari.internal_utilities import aio
from hikari.internal_utilities import assertions
from hikari.internal_utilities import cache
from hikari.internal_utilities import containers
from hikari.internal_utilities import dates
from hikari.internal_utilities import loggers
from hikari.internal_utilities import singleton_meta
from hikari.internal_utilities import storage
from hikari.internal_utilities import transformations



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
__all__ = ["GatewayBot", "GatewayActivity"]

import datetime
import typing

from hikari.core import entities
from hikari.internal_utilities import marshaller


@marshaller.attrs(slots=True)
class SessionStartLimit(entities.HikariEntity, entities.Deserializable):
    """Used to represent information about the current session start limits."""

    #: The total number of session starts the current bot is allowed.
    #:
    #: :type: :obj:`int`
    total: int = marshaller.attrib(deserializer=int)

    #: The remaining number of session starts this bot has.
    #:
    #: :type: :obj:`int`
    remaining: int = marshaller.attrib(deserializer=int)

    #: The timedelta of when :attr:`remaining` will reset back to :attr:`total`
    #: for the current bot.
    #:
    #: :type: :obj:`datetime.timedelta`
    reset_after: datetime.timedelta = marshaller.attrib(
        deserializer=lambda after: datetime.timedelta(milliseconds=after),
    )


@marshaller.attrs(slots=True)
class GatewayBot(entities.HikariEntity, entities.Deserializable):
    """Used to represent gateway information for the connected bot."""

    #: The WSS URL that can be used for connecting to the gateway.
    #:
    #: :type: :obj:`str`
    url: str = marshaller.attrib(deserializer=str)

    #: The recommended number of shards to use when connecting to the gateway.
    #:
    #: :type: :obj:`int`
    shard_count: int = marshaller.attrib(raw_name="shards", deserializer=int)

    #: Information about the bot's current session start limit.
    #:
    #: :type: :obj:`SessionStartLimit`
    session_start_limit: int = marshaller.attrib(deserializer=SessionStartLimit.deserialize)


@marshaller.attrs()
class GatewayActivity(entities.Deserializable, entities.Serializable):
    #: The activity name.
    #:
    #: :type: :obj:`str`
    name: str = marshaller.attrib(deserializer=str, serializer=str)

    #: The activity url. Only valid for ``STREAMING`` activities.
    #:
    #: :type: :obj:`str`, optional
    url: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str, if_none=None, if_undefined=None)

    # TODO: implement enum for this.
    #: The activity type.
    #:
    #: :type: :obj:`int`
    type: int = marshaller.attrib(deserializer=int, serializer=int, if_undefined=0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Anonymous system information that we have to provide to Discord when using their API.

This information contains details such as the version of Python you are using, and
the version of this library, the OS you are making requests from, etc.

This information is provided to enable Discord to detect that you are using a
valid bot and not attempting to abuse the API.
"""
__all__ = ["library_version", "python_version", "system_type", "user_agent"]

import platform

from hikari.internal_utilities import cache


@cache.cached_function()
def library_version() -> str:
    """The version of the library being used.

    Returns
    -------
    :obj:`str`
        A string representing the version of this library.

    Example
    -------
    .. code-block:: python
    
        >>> from hikari.net import user_agent
        >>> print(user_agent.library_version())
        hikari 0.0.71
    """
    from hikari._about import __version__

    return f"hikari {__version__}"


@cache.cached_function()
def python_version() -> str:
    """The python version being used.

    Returns
    -------
    :obj:`str`
        A string representing the version of this release of Python.

    Example
    -------
    .. code-block:: python

        >>> from hikari.net import user_agent
        >>> print(user_agent.python_version())
        CPython 3.8.1 GCC 9.2.0
    """
    attrs = [
        platform.python_implementation(),
        platform.python_version(),
        platform.python_branch(),
        platform.python_compiler(),
    ]
    return " ".join(a.strip() for a in attrs if a.strip())


@cache.cached_function()
def system_type() -> str:
    """The operating system being used.

    Returns
    -------
    :obj:`str`
        A string representing the system being used.

    Example
    -------
    .. code-block:: python

        >>> from hikari.net import user_agent
        >>> print(user_agent.system_type())
        Linux-5.4.15-2-MANJARO-x86_64-with-glibc2.2.5
    """
    # Might change this eventually to be more detailed, who knows.
    return platform.platform()


@cache.cached_function()
def user_agent() -> str:
    """The user agent of the bot

    Returns
    -------
    :obj:`str`
        The string to use for the library ``User-Agent`` HTTP header that is required
        to be sent with every HTTP request.

    Example
    -------
    .. code-block:: python

        >>> from hikari.net import user_agent
        >>> print(user_agent.user_agent())
        DiscordBot (https://gitlab.com/nekokatt/hikari, 0.0.71) CPython 3.8.1 GCC 9.2.0 Linux
    """
    from hikari._about import __version__, __url__

    system = system_type()
    python = python_version()
    return f"DiscordBot ({__url__}, {__version__}) {python} {system}"
